"""
Excel Add-on
2021
Babbage B.V., Falko Giepmans <falkogiepmans@babbage.com>
"""

# importing python modules
import uuid
import wx
import time
import os
import re
from comtypes import COMError

# importing NVDA modules
import api
import appModuleHandler
import scriptHandler
import ui
import config
import gui
import addonHandler
import eventHandler
import string
from keyboardHandler import KeyboardInputGesture
from logHandler import log
from scriptHandler import script
from NVDAObjects.window.excel import ExcelCell, \
	ExcelWorksheet, xlUp, xlDown

# importing our own modules
from . import excelElementsList
from . import monitorCells
import time

addonHandler.initTranslation()

class loadingDialog(wx.Dialog):
	def __init__(self, parent, title):
		super().__init__(parent, title = title)

		mainSizer = wx.BoxSizer(wx.VERTICAL)

		sHelper = gui.guiHelper.BoxSizerHelper(parent = self, orientation = wx.VERTICAL)
		sHelper.addItem(wx.StaticText(self, label="loading license"))


with loadingDialog(None, "loading license") as dlg:
	dlg.Show()
	dlg.Raise()
	try:

		from . import excelBase
		dlg.Destroy()
	except Exception as e:

		log.info(e)
		wx.CallAfter(gui.messageBox,f"No valid license for Excel Navigator Pro found on this system! \n If you want to continue with Excel, either disable the Excel Navigator Pro addon or give a valid license in the Babbage License Manager.\n \n {e}")
		from . import excelBaseFake as excelBase



# lists used for first character bug
gestureList =  ["kb:" + ii for ii in string.printable[:-6]]
gestureList = gestureList + [ "kb:numpad" + str(ii) for ii in range(0,10)] + [ "kb:numLockNumpad" + str(ii) for ii in range(0,10)]

toBeEnteredGestures = []

class AppModule(appModuleHandler.AppModule):
	"""
	The appModule overlays a couple of NVDA objects.
	"""
	def chooseNVDAObjectOverlayClasses(self, obj, clsList):
		log.info("chosing nvda object for appmoduel")
		if ExcelCell in clsList: # all excel cells
			clsList.insert(0, ExcelCellOverlay)
		if ExcelWorksheet in clsList: # all excel worksheets
			clsList.insert(0, ExcelWorksheetOverlay)

		
	@script(
		gesture="kb:NVDA+h",
		description = _("Open help document")
	)
	def script_openHelp(self, gesture):
		for addon in addonHandler.getAvailableAddons():
			if addon.name == "ExcelNavigatorPro":
				filedir = addon.getDocFilePath()
				os.startfile(filedir)

from gui import guiHelper

class MyDialog(wx.Dialog):
	def __init__(self, parent, title):
		super(MyDialog, self).__init__(parent, title = title)
		panel = wx.Panel()
		mainSizer = wx.BoxSizer(wx.VERTICAL)
		sHelper = guiHelper.BoxSizerHelper(parent = self, orientation=  wx.VERTICAL)

		self.checkboxRow =sHelper.addItem(wx.CheckBox(self, label = _("row header")))
		# self.minrowNumber = sHelper.addLabeledControl("min row nr", wx.SpinCtrl)
		# self.maxrowNumber = sHelper.addLabeledControl("max row nr", wx.SpinCtrl)

		self.checkboxCol =sHelper.addItem(wx.CheckBox(self, label=_("column header") ))

		# self.mincolNumber = sHelper.addLabeledControl("min col nr", wx.SpinCtrl)
		# self.maxcolNumber = sHelper.addLabeledControl("max col nr", wx.SpinCtrl)

		sHelper.addDialogDismissButtons(wx.OK | wx.CANCEL)
		mainSizer.Add(sHelper.sizer, border=guiHelper.BORDER_FOR_DIALOGS, flag=wx.ALL)
		mainSizer.Fit(self)
		self.SetSizer(mainSizer)
		self.Layout()
	def getValues(self):
		values = []
		values.append( self.checkboxRow.GetValue() )
		# values.append(self.minrowNumber.GetValue())
		# values.append(self.maxrowNumber.GetValue())

		values.append(self.checkboxCol.GetValue())
		# values.append(self.mincolNumber.GetValue())
		# values.append(self.maxcolNumber.GetValue())
		return values

def runScriptModalDialog(dialog, callback=None):
	"""Run a modal dialog from a script.
	This will not block the caller,
	but will instead call C{callback} (if provided) with the result from the dialog.
	The dialog will be destroyed once the callback has returned.
	@param dialog: The dialog to show.
	@type dialog: C{wx.Dialog}
	@param callback: The optional callable to call with the result from the dialog.
	@type callback: callable
	"""
	def run():
		gui.mainFrame.prePopup()
		res = dialog.ShowModal()
		gui.mainFrame.postPopup()
		if res == wx.ID_OK:
			values = dialog.getValues()
			if callback:
				callback(values)
		dialog.Destroy()
	wx.CallAfter(run)


class ExcelCellOverlay(excelBase.ExcelCellOverlay):
	"""
	Overlay class for the excelCell objects. Adds aditional scripts
	"""
	# sets custom TI as TI
	treeInterceptorClass = excelElementsList.ExcelBrowseModeTreeInterceptorOverlay




	@script(
		# Translators: The description for a script in the Excel App-Module
		description=_("reports row header of the current cell"),
		gesture="kb:NVDA+control+shift+r")
	def script_reportRowHeader(self, gesture):
		ui.message(f"{self._get_rowHeaderText()}")

	@script(
		# Translators: The description for a script in the Excel App-Module
		description=_("reports column header of the current cell"),
		gesture="kb:NVDA+control+shift+c")
	def script_reportColumnHeader(self, gesture):
		ui.message(f"{self._get_columnHeaderText()}")

	@script(
		# Translators: The description for a script in the Excel App-Module
		description=_("reports locattion of the current cell"),
		gesture="kb:NVDA+shift+p")
	def script_reportCoordinates(self, gesture):
		ui.message(f"{self._get_cellCoordsText()}")

	@script(
		# Translators: The description for a script in the Excel App-Module
		description=_("reports formula of the current cell"),
		gesture="kb:NVDA+shift+f")
	def script_reportFormula(self, gesture):
		formula = self.excelCellInfo.formula
		if formula:
			ui.message(f"{formula}")
		else:
			# Translators: The description for a script in the Excel App-Module
			ui.message(_("{cellCoords} contains no formula").format(cellCoords = self._get_cellCoordsText()) )

	@script(
		description=_("Go back to last cell before monitor cell"),
		gesture = "kb:NVDA+shift+backspace"
	)
	def script_goBackToCellBeforeMonitorCell(self, gesture):
		self.goBackToCellBeforeMonitorCell()

	@script(
		gesture = "kb:NVDA+shift+h"
	)
	def script_headerDialog(self, gesture):
		dlg = MyDialog(None, _("Set as overall header cell"))
		runScriptModalDialog(dlg, self.dialogReturned)


	@script(
		# Translators: The description for a script in the Excel App-Module
		description=_("set cell as monitor cell"),
		gestures=["kb:NVDA+control+1",
				  "kb:NVDA+control+2",
				  "kb:NVDA+control+3",
				  "kb:NVDA+control+4",
				  "kb:NVDA+control+5",
				  "kb:NVDA+control+6",
				  "kb:NVDA+control+7",
				  "kb:NVDA+control+8",
				  "kb:NVDA+control+9",
				  "kb:NVDA+control+0",
				  "kb:NVDA+control+numpad1",
				  "kb:NVDA+control+numpad2",
				  "kb:NVDA+control+numpad3",
				  "kb:NVDA+control+numpad4",
				  "kb:NVDA+control+numpad5",
				  "kb:NVDA+control+numpad6",
				  "kb:NVDA+control+numpad7",
				  "kb:NVDA+control+numpad8",
				  "kb:NVDA+control+numpad9",
				  "kb:NVDA+control+numpad0"
				  ]
	)
	def script_setMonitorCell(self, gesture):
		# take index from gesture name
		monitorIndex = int(gesture._get_displayName()[-1])
		scriptCount = scriptHandler.getLastScriptRepeatCount()

		# depending on scriptCount, either add or remove the monitor cell
		if scriptCount == 0:
			self.parent.setAsMonitorCell(self, monitorIndex)
		elif scriptCount == 1:
			self.parent.forgetMonitorCell(monitorIndex )

	""" 
	The next 10 scripts are essentially the same except for the used gesture, the reason why these functions are not
	combined is because of the second press. this second press of a certain gesture allows the user to go to the selected
	monitor cell. If the below would be condensed into 1 script, a fast press of {NVDA+Shift+1} and {NVDA+shift+3} would result 
	in moving to monitor cell 3. Alternatives, like using {NVDA+ctrl+number} or {NVDA+Alt+number} to remove a 
	monitor cell might be better but do not work. NVDA+control+4 does not work because control+4 is the underline shortkey
	and NVDA+alt+8 does not work for an unknown reason. 
	"""
	# Translators: The description for a script in the Excel App-Module
	descriptionString = _("Go to specified monitor cell")
	@script(
		# Translators: The description for a script in the Excel App-Module
		description=descriptionString,
		gestures=["kb:NVDA+shift+1",
				  "kb:NVDA+shift+numpad1"]
	)
	def script_specificMonitorCellAction1(self, gesture):
		self.specificMonitorCellAction(gesture)
	@script(
		# Translators: The description for a script in the Excel App-Module
		description=descriptionString,
		gestures=["kb:NVDA+shift+2",
				  "kb:NVDA+shift+numpad2"]
	)
	def script_specificMonitorCellAction2(self, gesture):
		self.specificMonitorCellAction(gesture)
	@script(
		# Translators: The description for a script in the Excel App-Module
		description=descriptionString,
		gestures=["kb:NVDA+shift+3",
				  "kb:NVDA+shift+numpad3"]
	)
	def script_specificMonitorCellAction3(self, gesture):
		self.specificMonitorCellAction(gesture)
	@script(
		# Translators: The description for a script in the Excel App-Module
		description=descriptionString,
		gestures=["kb:NVDA+shift+4",
				  "kb:NVDA+shift+numpad4"]
	)
	def script_specificMonitorCellAction4(self, gesture):
		self.specificMonitorCellAction(gesture)
	@script(
		# Translators: The description for a script in the Excel App-Module
		description=descriptionString,
		gestures=["kb:NVDA+shift+5",
				  "kb:NVDA+shift+numpad5"]
	)
	def script_specificMonitorCellAction5(self, gesture):
		self.specificMonitorCellAction(gesture)
	@script(
		# Translators: The description for a script in the Excel App-Module
		description=descriptionString,
		gestures=["kb:NVDA+shift+6",
				  "kb:NVDA+shift+numpad6"]
	)
	def script_specificMonitorCellAction6(self, gesture):
		self.specificMonitorCellAction(gesture)
	@script(
		# Translators: The description for a script in the Excel App-Module
		description=descriptionString,
		gestures=["kb:NVDA+shift+7",
				  "kb:NVDA+shift+numpad7"]
	)
	def script_specificMonitorCellAction7(self, gesture):
		self.specificMonitorCellAction(gesture)
	@script(
		# Translators: The description for a script in the Excel App-Module
		description=descriptionString,
		gestures=["kb:NVDA+shift+8",
				  "kb:NVDA+shift+numpad8"]
	)
	def script_specificMonitorCellAction8(self, gesture):
		self.specificMonitorCellAction(gesture)
	@script(
		# Translators: The description for a script in the Excel App-Module
		description=descriptionString,
		gestures=["kb:NVDA+shift+9",
				  "kb:NVDA+shift+numpad9"]
	)
	def script_specificMonitorCellAction9(self, gesture):
		self.specificMonitorCellAction(gesture)
	@script(
		# Translators: The description for a script in the Excel App-Module
		description=descriptionString,
		gestures=["kb:NVDA+shift+0",
				  "kb:NVDA+shift+numpad0"]
	)
	def script_specificMonitorCellAction0(self, gesture):
		self.specificMonitorCellAction(gesture)


	def specificMonitorCellAction(self, gesture ):
		""" Depending on the amount of presses and the gesture: """
		scriptCount = scriptHandler.getLastScriptRepeatCount()
		# get index from gesture name
		monitorIndex = int(gesture._get_displayName()[-1])
		monitorCell = self.parent.monitorCellTracker.getMonitorCellInfoAtIndex(monitorIndex)

		if not monitorCell:
			# Translators: Message in Excel
			ui.message(_("No monitor cell at index {monitorIndex}").format(monitorIndex = monitorIndex))
			return

		# depending on scriptCount, report or go-to monitor cell
		if scriptCount == 0:
			self.reportSpecificMonitorCell(monitorCell)
		elif scriptCount == 1:
			# set last cell in case the user wants to jump back
			self.parent.monitorCellTracker.setLastCell(self.rowNumber, self.columnNumber)
			self.goToMonitorCell(monitorCell)


	def reportSpecificMonitorCell(self, monitorCell):
		monitorCellObject = self.parent.excelWorksheetObject.cells(monitorCell[0], monitorCell[1])

		value = monitorCellObject.value()
		if not value == None:
			ui.message(f"{value}")
		else:
			# Translators: Message in a script in Excel
			ui.message(_("{cellCoords} has no value").format(cellCoords = excelElementsList.getCoordsFromCellObject(monitorCellObject) ))

	def goToMonitorCell(self, monitorCell):
		super().goToMonitorCell(monitorCell)


class ExcelWorksheetOverlay(excelBase.ExcelWorksheetOverlay):
	treeInterceptorClass = excelElementsList.ExcelBrowseModeTreeInterceptorOverlay



	# this is also defined in window/excel.py of NVDA but is redefined here so we can encoperate maxRow and maxCol,
	# also, id is added which can return the uuid of the header.
	re_definedName=re.compile(
		u'^((?P<sheet>(\'[^\']+\'|[^!]+))!)?'
		u'(?P<name>\w+)'
		u'\.(?P<maxRow>[0-9]+)?\.(?P<maxCol>[0-9]+)?\.(?P<id>[0-9a-f]+)$'
	)
	#
	# this function is overwritten to be able to extract maxRow and maxCol from the cell name.
	def populateHeaderCellTrackerFromNames(self,headerCellTracker):
		sheetName=self.excelWorksheetObject.name
		for x in self.excelWorksheetObject.parent.names:
			fullName=x.name
			nameMatch=self.re_definedName.match(fullName)
			if not nameMatch:
				continue
			sheet=nameMatch.group('sheet')
			if sheet and sheet[0]=="'" and sheet[-1]=="'":
				sheet=sheet[1:-1]
			if sheet and sheet!=sheetName:
				continue
			name=nameMatch.group('name').lower()
			isColumnHeader=isRowHeader=False
			if name.startswith('title'):
				isColumnHeader=isRowHeader=True
			elif name.startswith('columntitle'):
				isColumnHeader=True
			elif name.startswith('rowtitle'):
				isRowHeader=True
			else:
				continue
			try:
				headerCell=x.refersToRange
			except COMError:
				continue
			if headerCell.parent.name!=sheetName:
				continue
			minColumnNumber=maxColumnNumber=minRowNumber=maxRowNumber=None
			maxRowNumber=int( nameMatch.group('maxRow') )
			maxColumnNumber=int( nameMatch.group('maxCol') )
			id = nameMatch.group("id")
			log.info(id)
			log.info(f"{maxColumnNumber} { maxRowNumber}")
			if maxColumnNumber is None:
				maxColumnNumber=self._getMaxColumnNumberForHeaderCell(headerCell)
			if maxColumnNumber == 0:
				maxColumnNumber = None
			if maxRowNumber == 0:
				maxRowNumber = None

			self.headerCellTracker.addHeaderCellInfo(rowNumber=headerCell.row,columnNumber=headerCell.column,rowSpan=headerCell.rows.count,colSpan=headerCell.columns.count,minRowNumber=minRowNumber,maxRowNumber=maxRowNumber,minColumnNumber=minColumnNumber,maxColumnNumber=maxColumnNumber,name=fullName,isColumnHeader=isColumnHeader,isRowHeader=isRowHeader)


	def _get_monitorCellTracker(self):
		""" Based on _get_headerCellTracker"""
		self.monitorCellTracker = monitorCells.MonitorCellTracker()
		self.populateMonitorCellTrackerFromNames()
		return self.monitorCellTracker


	def populateMonitorCellTrackerFromNames(self):
		monitorCellList = monitorCells.collectMonitorCellsFromWorksheet(self.excelWorksheetObject)
		for monitorTuple in monitorCellList:
			self.monitorCellTracker.addMonitorCellInfo(rowNumber=monitorTuple[0], columnNumber=monitorTuple[1], index=monitorTuple[2], name=monitorTuple[3])

	def setAsMonitorCell(self, cell, monitorIndex):

		sentences = [ _("Cell {cellCoords} is already a monitor cell."),
					  _("Monitor cell index {monitorIndex} is already in use"),
					  _("set {cellCoords} as monitor cell {monitorIndex}") ]

		super().setAsMonitorCellBase(cell, monitorIndex, sentences)


	def forgetMonitorCell(self, monitorIndex):
		""" based on forgetHeaderCell"""
		(row, column) = self.monitorCellTracker.DictByIndex[monitorIndex]
		info=self.monitorCellTracker.getMonitorCellInfoAt(row, column)
		if not info:
			return False
		self.monitorCellTracker.removeMonitorCellInfo(info)
		self.excelWorksheetObject.parent.names(info.name).delete()
		ui.message(_("removed monitor cell {monitorIndex}").format(monitorIndex = monitorIndex))
		return True

