
# import NVDA modules
import browseMode
import api
import eventHandler
import addonHandler
from comtypes import COMError
from NVDAObjects.window import excel

# import our own modules
from . import monitorCells


addonHandler.initTranslation()


def getCoordsFromCellObject(excelCellObject):
    rawAddress = excelCellObject.address(False, False, 1, False)
    coords = rawAddress.split('!')[-1].split(':')
    if len(coords) == 2:
        # Translators: Used to express an address range in excel.
        return _("{start} through {end}").format(start=coords[0], end=coords[1])
    else:
        return coords[0]


class ExcelBrowseModeTreeInterceptorOverlay(excel.ExcelBrowseModeTreeInterceptor):
    """
    The new tree interecptor that should be used with the overlay classes. This TI adds the capability to add monitor
    cells as well as datablocks to the ElementsList (NVDA+F7)
    """

    def _get_ElementsListDialog(self):
        # _get_ElementsListDialog is overwritten from parent so that monitor cells and datablocks are included
        return ElementsListDialog

    def _iterNodesByType(self, nodeType, direction="next", pos=None):
        # _iterNodesByType is extended to add monitor cells and blocks to elements list
        if nodeType == "monitor":
            return MonitorExcelCollectionQuicknavIterator(nodeType, self.rootNVDAObject.excelWorksheetObject, direction,
                                                          None).iterate()
        elif nodeType == "block":
            return dataBlockExcelCollectionQuicknavIterator(nodeType, self.rootNVDAObject.excelWorksheetObject,
                                                            direction, None).iterate()
        else:
            return super()._iterNodesByType(nodeType, direction, pos)


class ExcelMonitorQuickNavItem(excel.ExcelQuickNavItem):
    """
    Adjusted navItem used in ElementsList for monitor cells
    """

    def __init__(self, nodeType, document, monitorInfo, items):
        self.document = document
        self.monitorInfo = monitorInfo
        self.row = self.monitorInfo[0]
        self.column = self.monitorInfo[1]
        self.index = self.monitorInfo[2]
        self.monitorCollection = None
        self.cellObject = self.document.cells(self.row, self.column)
        self.label = _("Monitor Cell {index} at {cellCoords}: \t {value}").format(  index = self.index,
                                                                                    cellCoords = getCoordsFromCellObject(self.cellObject),
                                                                                    value = self.cellObject.value() )
        super().__init__(nodeType, document, self.cellObject, self.monitorCollection)

    def __lt__(self, other):
        return self.index < other.index

    def moveTo(self):
        self.cellObject.Select()
        self.cellObject.Activate()
        eventHandler.queueEvent("gainFocus", api.getDesktopObject().objectWithFocus())


class MonitorExcelCollectionQuicknavIterator(excel.ExcelQuicknavIterator):
    """
    Allows iterating over a MS excel Sheets collection emitting L{QuickNavItem} object.
    """
    quickNavItemClass = ExcelMonitorQuickNavItem

    def collectionFromWorksheet(self, worksheetObject):
        # used in the corresponding iterator.
        try:
            return monitorCells.collectMonitorCellsFromWorksheet(worksheetObject)
        except COMError:
            return None


class ExcelDataBlockQuickNavItem(excel.ExcelQuickNavItem):
    """
    Adjusted navItem used in ElementsList for datablocks, i.e. all smallest squares encompassing all connected
    parts of data
    """

    def __init__(self, nodeType, document, BlockInfo, items):
        self.document = document
        self.BlockInfo = BlockInfo
        self.row = self.BlockInfo[0]
        self.column = self.BlockInfo[1]
        self.height = self.BlockInfo[2]
        self.width = self.BlockInfo[3]
        self.size = self.height * self.width
        self.dataBlokCollection = None
        self.cellObject = self.document.cells(self.row, self.column)

        # Translators: Label of a data block in excel for ElementsList.
        self.label = _("Datablock: \t {h} rows by {w} columns  \t\t at {cellCoords}").format(h = self.height, w = self.width, cellCoords = getCoordsFromCellObject(self.cellObject))
        super().__init__(nodeType, document, self.cellObject, self.dataBlokCollection)

    def __lt__(self, other):
        return self.size < other.size

    def moveTo(self):
        self.cellObject.Select()
        self.cellObject.Activate()
        eventHandler.queueEvent("gainFocus", api.getDesktopObject().objectWithFocus())


class dataBlockExcelCollectionQuicknavIterator(excel.ExcelQuicknavIterator):
    """
    Allows iterating over an MS excel Sheets collection emitting L{QuickNavItem} object.
    """
    quickNavItemClass = ExcelDataBlockQuickNavItem

    def collectDataBlockCellsFromWorksheet(self, excelWorksheetObject):
        regionList = []
        for area in excelWorksheetObject.cells.specialCells(excel.xlCellTypeConstants).Areas:
            cell = area.cells.item(1)
            region = cell.currentRegion
            regionTuple = (region.row, region.column, region.rows.count, region.columns.count)
            regionList.append(regionTuple)

        regionList = list(set(regionList))
        regionList = sorted(regionList, key=lambda x: x[2] * x[3], reverse=True)
        return regionList

    def collectionFromWorksheet(self, worksheetObject):
        try:
            return self.collectDataBlockCellsFromWorksheet(worksheetObject)
        except(COMError):
            return None


class ElementsListDialog(browseMode.ElementsListDialog):
    helpId = "ExcelElementsList"

    ELEMENT_TYPES = (
        # Translators: The label of a radio button to select the type of element
        # in the browse mode Elements List dialog.
        ("chart", _("&Charts")),
        # Translators: The label of a radio button to select the type of element
        # in the browse mode Elements List dialog.
        ("comment", _("N&otes")),
        # Translators: The label of a radio button to select the type of element
        # in the browse mode Elements List dialog.
        ("formula", _("Fo&rmulas")),
        # Translators: The label of a radio button to select the type of element
        # in the browse mode Elements List dialog.
        ("formField", _("&Form fields")),
        # Translators: The label of a radio button to select the type of element
        # in the browse mode Elements List dialog.
        ("monitor", _("&Monitor Cells")),
        # Translators: The label of a radio button to select the type of element
        # in the browse mode Elements List dialog.
        # ("block", _("&Data Blocks")),
        # # Translators: The label of a radio button to select the type of element
        # # in the browse mode Elements List dialog.
        ("sheet", _("&Sheets")),
    )
