
from ctypes import *
from ctypes.wintypes import *
import os
import wx # type: ignore
import addonHandler
import gui
import core
import queueHandler
import watchdog
from logHandler import log
import config
import datetime
addonHandler.initTranslation()

from . import licenseExtension
# This is a dummy function so the _() function of i18n is not flagged as undefined
def _(message):
    return message

a4uLicenseMessageMapping = {
    "to verify there are activations remaining, and the API Key and Product ID are correct.": _("This license key is not valid for this product."), # type: ignore
    "customer account does not exist for this API Key.": _("This license key is not valid."), # type: ignore
    "not activate API Key. No API resources available.": _("No activations remaining.") # type: ignore
}
# something can go wrong with the conversion in the dll when tranlating the json return to some string class in c++,
# this causes that in the beginning of the error message some chinese characters appear, thus we check if the message
# ends with a known message and return the correct message
def getMessageEndWithA4uLicenseMessage(message):
    for key in a4uLicenseMessageMapping:
        if message.endswith(key):
            return a4uLicenseMessageMapping[key]
    return None

# load lib modules
addon: addonHandler.Addon = addonHandler.getCodeAddon()


def ActivateLicense(licenseKey):
    returnCode = licenseExtension.activateProduct(licenseKey)
    return returnCode == licenseExtension.a4u_success

def DeactivateLicense():
    returnCode = licenseExtension.deactivateProduct()
    return returnCode == licenseExtension.a4u_success

class LicenseDialog(wx.Dialog):

    def restartNVDA(self):
        
        queueHandler.queueFunction(queueHandler.eventQueue, core.restart)

    def addTopMessage(self):
        addonManifest = addonHandler.getCodeAddon().manifest
        currentVersion = addonManifest['version']
        if self.isLicensed:
            log.debugWarning("Excel Navigator Pro is licensed")
            message = _("Your copy of Excel Navigator Pro {currentVersion} is properly licensed.").format(currentVersion=currentVersion)
        else:
            log.debugWarning("Excel Navigator Pro not licensed")
            self.shouldShowAtStart = True
            message = _("Your copy of Excel Navigator Pro {currentVersion} doesn't seem to be licensed.").format(currentVersion=currentVersion)       

        self.main_sizer.Add(wx.StaticText(self, label=message),  wx.SizerFlags(0).Align(wx.LEFT).Border(wx.ALL, 5))

    def addLicenseKeyOrDeactivateButton(self):
        #Translators: The input field to enter the license key
        if not self.isLicensed:
            activationButton = self.main_sizer_helper.addItem(wx.Button(self, label=_("activate with key")), flag = wx.EXPAND)
            activationButton.Bind(wx.EVT_BUTTON, self.activate)
        else:
            deActivationButton = self.main_sizer_helper.addItem(wx.Button(self, label=_("deactivate key")), flag = wx.EXPAND)
            deActivationButton.Bind(wx.EVT_BUTTON, self.deactivate)

    def addButtons(self):
        offlineActivation = self.main_sizer_helper.addItem(wx.Button(self, label=_("Offline activation")), flag = wx.EXPAND)
        offlineActivation.Bind(wx.EVT_BUTTON, self.offlineActivation)            
        bHelper = self.main_sizer_helper.addDialogDismissButtons(gui.guiHelper.ButtonHelper(wx.HORIZONTAL))
        closeButton = bHelper.addButton(self, wx.ID_CLOSE, label=_("&Close"))
        closeButton.Bind(wx.EVT_BUTTON, self.onClose)
        closeButton.SetFocus()


    def __init__(self, parent):
        # Translators: Title of the license dialog
        windowStyle = wx.DEFAULT_DIALOG_STYLE | wx.CENTER
        super().__init__(parent, id=wx.ID_ANY, title=_("Manage Excel Navigator Pro License"), style = windowStyle)
        self.isLicensed = licenseExtension.isLicensed()
        self.isAutomaticActivation = False
        self.isOfflineActivation = False
        self.isDemo = False
        self.demoOver = False
        self.shouldShowAtStart = False
        self.demoDatabaseID = 0
        

        self.main_sizer = wx.BoxSizer(wx.VERTICAL)
        self.main_sizer_helper = gui.guiHelper.BoxSizerHelper(self, orientation=wx.VERTICAL)

        self.addTopMessage()
        self.addLicenseKeyOrDeactivateButton()
        self.addButtons()        

        self.main_sizer.Add(self.main_sizer_helper.sizer, border = gui.guiHelper.BORDER_FOR_DIALOGS, flag=wx.ALL)
        self.main_sizer.Fit(self)
        self.SetSizer(self.main_sizer)
        self.Center(wx.BOTH )

    def activate(self, evt):
        success = False
        with activationDialog(self) as dlg:
            dlg.ShowModal()
            success = dlg.success
        if success:
            self.Close()
            self.Destroy()  
            wx.CallAfter(self.restartNVDA)

    def deactivate(self, evt):
        try:
            success = watchdog.cancellableExecute(DeactivateLicense)
        except Exception as e:
            success = False
            message = _(f"There was a timeout while performing your request. {e}")
            log.error("Activation error", exc_info=True)
        if not success:
            message = _("Excel Navigator Pro could not be deactivated.")
            gui.messageBox(_("An error has occured:\n{error}").format(error=message),
                    _("Error"), wx.OK | wx.ICON_ERROR)
        else:
            gui.messageBox(_("Excel Navigator Pro has been deactivated! NVDA will be restarted."),
                    _("Congratulations!"), wx.OK | wx.ICON_EXCLAMATION)
        if success:
            self.Close()
            self.Destroy()
            wx.CallAfter(self.restartNVDA)

    def offlineActivation(self, evt):

        with offlineDialog(self) as dlg:
            dlg.ShowModal()

        self.Close()
        self.Destroy()

    
    def onClose(self, evt):
        self.Close()


class activationDialog(wx.Dialog):

    def __init__(self, parent):
        windowStyle = wx.DEFAULT_DIALOG_STYLE | wx.STAY_ON_TOP | wx.CENTER
        super().__init__(parent, id=wx.ID_ANY, title=_("Manage Excel Nav Pro License"), style = windowStyle)
        
        self.main_sizer = wx.BoxSizer(wx.VERTICAL)
        self.main_sizer_helper = gui.guiHelper.BoxSizerHelper(self, orientation=wx.VERTICAL)

        self.key = self.main_sizer_helper.addLabeledControl(_("&License Key:"), wx.TextCtrl)

        self.Bind(wx.EVT_BUTTON, self.on_ok, id=wx.ID_OK)
        self.main_sizer.Add(self.main_sizer_helper.sizer, border = gui.guiHelper.BORDER_FOR_DIALOGS, flag=wx.ALL)
        self.main_sizer_helper.addDialogDismissButtons(self.CreateButtonSizer( wx.OK| wx.CANCEL))
        self.main_sizer.Fit(self)
        self.SetSizer(self.main_sizer)
        self.Center(wx.BOTH )
        self.key.SetFocus()

    def on_ok(self, evt):
        self.activate()
    
    def activate(self):
        if not self.key.Value:
            gui.messageBox(_("You must enter a valid license key."), _("Error"), wx.OK | wx.ICON_ERROR)
            self.key.SetFocus()
            return

        try:
            self.success = watchdog.cancellableExecute(ActivateLicense, self.key.Value)
        except Exception as e:
            self.success = False
            message = _("There was an error while performing your request. \n{e}").format(e)
            log.error(_("Activation error"), exc_info=True)


        if not self.success:
            message = _("Excel Nav Pro could not be activated.")
            gui.messageBox(_("An error has occured:\n{error}").format(error=message),
                    _("Error"), wx.OK | wx.ICON_ERROR)
        else:
            gui.messageBox(_("Excel Navigator Pro has been activated! NVDA will be restarted."),
                    _("Congratulations!"), wx.OK | wx.ICON_EXCLAMATION)
            queueHandler.queueFunction(queueHandler.eventQueue, core.restart)
        if self.success:
            self.Close()


class offlineDialog(wx.Dialog):

    def __init__(self, parent):
        # Translators: Title of the license dialog
        super().__init__(parent, id=wx.ID_ANY, title=_("offline Activation"))
        self.isLicensed = licenseExtension.isLicensed()
        addonManifest = addonHandler.getCodeAddon().manifest
        currentVersion = addonManifest['version']

        message = _("We offer offline activations as well. For this you need to mail your licensekey and instance ID to support@babbage.com. \nWe will send you an unique activation hash so that Excel Nav Pro is licensed without the need of verifying online.")

        main_sizer = wx.BoxSizer(wx.VERTICAL)
        main_sizer.Add(wx.StaticText(self, label=message))
        main_sizer_helper = gui.guiHelper.BoxSizerHelper(self, orientation=wx.VERTICAL)
        #Translators: The input field to enter the license key


        productChoices = ["Excel Nav Pro"]
        self.productIDs = [licenseExtension.prodID]
        self.regPath = [licenseExtension.prodName]
        self.productID = main_sizer_helper.addLabeledControl("&license version", wx.Choice, choices=productChoices)
        self.productID.SetSelection(0)
        self.productID.Bind(wx.EVT_CHOICE, self.changedProduct)

        id = licenseExtension.getInstanceID(self.productIDs[self.productID.GetSelection()])

        self.key = main_sizer_helper.addLabeledControl(_("&Instance ID:"), wx.TextCtrl)
        self.key.Enabled = True
        self.key.Value = id
        self.lic = main_sizer_helper.addLabeledControl(_("&license key:"), wx.TextCtrl)
        self.lic.Enabled = True
        self.activationHash = main_sizer_helper.addLabeledControl(_("&Activation hash:"), wx.TextCtrl)
        self.activationHash.Enabled = True

        main_sizer_helper.addDialogDismissButtons(self.CreateButtonSizer(wx.OK | wx.CANCEL))
        self.Bind(wx.EVT_BUTTON, self.on_ok, id=wx.ID_OK)
        main_sizer.Add(main_sizer_helper.sizer, border = gui.guiHelper.BORDER_FOR_DIALOGS, flag=wx.ALL)
        main_sizer.Fit(self)
        self.SetSizer(main_sizer)
        self.Center(wx.BOTH )
        self.Show()
        self.key.SetFocus()


    def changedProduct(self, evt):
        id = licenseExtension.getInstanceID(self.productIDs[self.productID.GetSelection()])
        self.key.Value = id
        evt.Skip()
    def on_ok(self, evt):
        if len(self.activationHash.Value.split()) > 0:

            regPath = os.path.join("SOFTWARE", "ACCESS4U",self.regPath[self.productID.GetSelection()])
            licenseExtension.saveDataToRegistry(self.lic.Value, self.activationHash.Value, regPath)


        self.Close()
        self.Destroy()





